/** Represents a scope within a TableOfVars instance */
function TableOfVarsScope(
    parentTable,    // TableOfVarsAbstract instance containing this instance
    line,           // Line number of line starting this scope
    className       // Optional: Name of class to create (only if this is a scope to create an object
) {
    TableOfVars.parent.call(this, parentTable);
    this.line = line;
    if ((typeof className == 'string') || (className instanceof String))
        this.className = className;
    this.nodeData = parentTable.addTbody();
    this._buildFirstRow();
    if (!this.isObjectCreation)
        this._varThis = new VarTypeReference(this, 'this', true, 0);
}
TableOfVarsScope.parent = TableOfVarsAbstract;
TableOfVarsScope.prototype = Object.create(TableOfVarsScope.parent.prototype);
TableOfVarsScope.prototype.constructor = TableOfVarsScope;

/** Name of class to create (or null if there is no object to be created) */
TableOfVarsScope.prototype.className = null;

/** Is this scope used to create an object? */
Object.defineProperty(TableOfVarsScope.prototype, 'isObjectCreation', {
    enumerable: true,
    get() { return (this.className !== null); }
});

/** Line number of line starting this scope */
TableOfVarsScope.prototype.line = -1;

/**
* Add title row of scope ("this = …")
* @return undefined
*/
TableOfVarsScope.prototype._buildFirstRow = function() {
    const doc = this.getDocument();
    const tr = doc.createElement('tr');
    const th = ScoTools.createHtml('th', tr, { colspan: 3 }, this.isObjectCreation ? 'Erzeuge Objekt: ' : 'Methodenaufruf: ');
    const code = ScoTools.createHtml('code', th, { 'class': 'hljs language-java' }).innerHTML = this._requestformatedline(this.line);
    ScoTools.createHtml('button', ScoTools.createHtml('td', tr), { title: this.isObjectCreation ? 'Objekt speichern' : 'Methodenaufruf abschließen' }, this.isObjectCreation ? '💾' : '❌')
        .addEventListener('click', () => this.tryRemove());
    this.nodeData.appendChild(tr);
}

/**
* Try to remove this scope
* @return undefined
*/
TableOfVarsScope.prototype.tryRemove = function(
    suppressReturn      // Optional: Set true in order not to change active source code line
) {
    const topScope = this.scopes[this.scopes.length - 1];
    if (this != topScope) {
        const solange = topScope.isObjectCreation ? ('Solange das Objekt der Klasse „' + topScope.className + '“ nicht fertig erzeugt ist') : 'Solange der aktuellste Methodenaufruf nicht abgeschlossen ist';
        const kannnicht = this.isObjectCreation ? ' kann das Erzeugen dieses Objekts' : ' kann dieser Methodenaufruf';
        this._addfailed(solange + kannnicht + ' nicht abgeschlossen werden.');
        return;
    }
    const isNestedObjectCreation = ((this.scopes.length > 1) && this.scopes[this.scopes.length - 2].isObjectCreation);
    if (this.isObjectCreation && isNestedObjectCreation) {
        this.scopes[this.scopes.length - 2].addVars(this.toJSON().vars);
    }
    if (this.isObjectCreation && !isNestedObjectCreation) {
        const obj = this.createInstanceOfObject(this.className);
        if (obj instanceof TableOfVarsInstancelike)
            obj.load(this.toJSON().vars);
    }
    this.remove();
    suppressReturn || this._requestreturn(this.line);
}

/**
* Force remove this scope
* @return undefined
*/
TableOfVarsScope.prototype.remove = function() {
    this.removeScope(this, this.nodeData);
}

/**
* Function used to load value of "this"
* @return Boolean       Value accepted?
*/
TableOfVarsScope.prototype.loadThis = function(
    value       // New value of "this"
) {
    if (this.isObjectCreation)
        return false;
    this._varThis.value = value;
    return true;
}

/**
* Convert contained data to JSON
* @return Object    JSON representation of contained data
*/
TableOfVarsScope.prototype.toJSON = function() {
    const vars = TableOfVarsScope.parent.prototype.toJSON.call(this);
    if (!this.isObjectCreation)
        vars.unshift(this._varThis.toJSON());
    return { className: this.className, line: this.line, vars: vars };
}
